//just expand Uno to 32kB of slow RAM
//by the cost of 11 pins.
//MCbx 2018

//PIN CONNECTIONS:
//SHIFT REGISTER PINS MUST BE THE FOLLOWING:
// PIN 13 -> CLOCK (595 pin 11 and the next one's 10)
// PIN 11 -> MOSI (595 pin 14, the first one only)
// PIN 10 -> LATCH (595 pin 12)
// PIN 12 -> 62256 /WE PIN

//Latch pin (595 pin 12 and next one's 12)
//Pins for bits 0-5 are fixed to A0..A5
//This way we have faster access by just reading PORT.
//Pins for BIT 6 and BIT 7 are 8 and 9 respectively
//This cannot be DEFINED as I'm using bit manipulation here.


//initialize SRAM interface
void mem_init()
{
    //DATA I/O
  DDRC=0; //pins A0..A* to INPUTS
  DDRB &= ~(1<<0);
  DDRB &= ~(1<<1); //pins 8 and 9 as inputs too

  //Initialize SPI
  DDRB |= _BV(PINB2); //SS out
  DDRB |= _BV(PINB3);  //mosi out
  DDRB |= _BV(PINB5);  //clk out
  DDRB &= ~(_BV(PINB4)); //miso in
  PORTB |= (1<<2);
  
  SPCR = (1<<SPE)|(1<<MSTR); //spirupt disable, spi endble, msb1 clklow idle, sample on edge, clk/4 rate
  SPSR |= 0x01;  // set SPI2x

}

//shift data out of spi
inline void shift_spi(volatile uint16_t data)
{
  PORTB &= ~(1<<2); //Latch low
  SPDR = data>>8;                    // Start the transmission
  while (!(SPSR & (1<<SPIF)))     // Wait the end of the transmission
  {
  };
  SPDR = data&0xFF;                    // Start the transmission
  while (!(SPSR & (1<<SPIF)))     // Wait the end of the transmission
  {
  };
   PORTB |= (1<<2); //Latch high
  return ; 
  
}

//Read byte from memory
byte read_byte(unsigned int addr)
{
  //address (with WE high) in
 // bitSet(addr,15);
  shift_spi(addr);
  
  //read byte
  byte a=PINC;
  bitWrite(a,6,bitRead(PINB,0));
  bitWrite(a,7,bitRead(PINB,1));
  return a;
}

//Write data into memory
//WE is Q7 of second register.
void write_byte(unsigned int addr, byte data)
{
  bitSet(addr,15); //address (with WE high) in
   shift_spi(addr);
   
  //bang /WE - a bit faster version
  SPCR=0; //spi temporary off
  DDRB|=(1<<4); //B4 as o/p
  PORTB &= ~(1<<4); // low
  
  //pins to outputs with data
  DDRC=255;
  DDRB|=(1<<0);
  DDRB|=(1<<1);


  PORTC=data; //portc is A-pins. In Uno we don't have much choice.
  data=data>>6;
  bitSet(data,2); //Latch high! Other PORTB pins are nevertheless driven by SPI cntrlr.
  PORTB=data;

 /* 
  //Bang the WE - address with WE low in - slow version
  bitClear(addr,15);
  shift_spi(addr);
  //address with WE  high in
  bitSet(addr,15);
  shift_spi(addr);
  */

  // /WE high again
  PORTB |= (1<<4); // high
  SPCR = (1<<SPE)|(1<<MSTR); //spi on
  //According to spec, /WE has to be internally pulled up. Thus, we can just turn
  //pin 12 to input, and it'll maintain high state.
  
  //data to inputs. They should ALWAYS be inputs 
  //and WE should be high if not in one writing moment.
  DDRC=0; 
  DDRB &= ~(1<<0);
  DDRB &= ~(1<<1);
}




void setup() {

  mem_init();
  
  Serial.begin(19200);
}

unsigned int rcursor=0;

void loop() {
 
//benchmark 
unsigned long m=millis();
  for (unsigned int i=0;i<32767;i++)
  {
  //  write_byte(i,255-i%255); //uncomment for memtesting
  write_byte(i,125); //comment for memtesting
  }
  m=millis()-m;
  Serial.print(m);
  Serial.print(" R:");
  m=millis();
  byte l=0;
  for (unsigned int i=0;i<32767;i++)
  {
    l=read_byte(i);
  //  if (l!=255-i%255) //uncomment for memtesting
  //  {
  //    Serial.println("E!");
  //  }
  }
  m=millis()-m;
  Serial.println(m);



/*

//Repeater of string - repeats at ^ sign
if (Serial.available())
{
  byte a=Serial.read();
  write_byte(rcursor,a);
  rcursor++;
  if (a=='^')
  {
    for (unsigned int u=0;u<rcursor;u++)
    {
      Serial.print((char)read_byte(u));
    }
  }
}
*/

  
/*  for (int i=0;i<256;i++)
  {
    write_byte(50,i);
    write_byte(32766,i+2);
    Serial.print(i);
    Serial.print (": ");
    Serial.print(read_byte(32766));
    Serial.print(" ");
    Serial.println(read_byte(50));
  }*/

//delay(500);
}